@@ -1,4 +1,6 @@
-from rl_toolkit.networks import Actor, Critic
+from rl_toolkit.networks import Actor, TwinCritic
+from rl_toolkit.policy import Policy
+from rl_toolkit.utils import VariableContainer
 
 import os
 import reverb
@@ -7,15 +9,15 @@ import wandb
 import tensorflow as tf
 
 
-class Learner:
+class Learner(Policy):
     """
     Learner (based on Soft Actor-Critic)
     =================
-    Paper: https://arxiv.org/pdf/1812.05905.pdf
+
     Attributes:
         env: the instance of environment object
         max_steps (int): maximum number of interactions do in environment
-        learning_starts (int): number of interactions before using policy network
+        warmup_steps (int): number of interactions before using policy network
         buffer_capacity (int): the capacity of experiences replay buffer
         batch_size (int): size of mini-batch used for training
         actor_learning_rate (float): learning rate for actor's optimizer
@@ -23,12 +25,13 @@ class Learner:
         alpha_learning_rate (float): learning rate for alpha's optimizer
         tau (float): the soft update coefficient for target networks
         gamma (float): the discount factor
-        model_a_path (str): path to the actor's model
-        model_c1_path (str): path to the critic_1's model
-        model_c2_path (str): path to the critic_2's model
+        actor_path (str): path to the actor's model
+        critic_path (str): path to the critic's model
+        db_path (str): path to the database checkpoint
         save_path (str): path to the models for saving
-        db_path (str): path to the database
         log_wandb (bool): log into WanDB cloud
+
+    Paper: https://arxiv.org/pdf/1812.05905.pdf
     """
 
     def __init__(
@@ -36,7 +39,7 @@ class Learner:
         # ---
         env,
         max_steps: int,
-        learning_starts: int = 10000,
+        warmup_steps: int = 10000,
         # ---
         buffer_capacity: int = 1000000,
         batch_size: int = 256,
@@ -48,22 +51,21 @@ class Learner:
         tau: float = 0.01,
         gamma: float = 0.99,
         # ---
-        model_a_path: str = None,
-        model_c1_path: str = None,
-        model_c2_path: str = None,
-        save_path: str = None,
+        actor_path: str = None,
+        critic_path: str = None,
         db_path: str = None,
+        save_path: str = None,
         # ---
         log_wandb: bool = False,
         log_interval: int = 64,
     ):
-        self._env = env
+        super(Learner, self).__init__(env, log_wandb)
+
         self._max_steps = max_steps
-        self._learning_starts = learning_starts
+        self._warmup_steps = warmup_steps
         self._gamma = tf.constant(gamma)
         self._tau = tf.constant(tau)
         self._save_path = save_path
-        self._log_wandb = log_wandb
         self._log_interval = log_interval
 
         # init param 'alpha' - Lagrangian constraint
@@ -81,76 +83,29 @@ class Learner:
             state_shape=self._env.observation_space.shape,
             action_shape=self._env.action_space.shape,
             learning_rate=actor_learning_rate,
-            model_path=model_a_path,
+            model_path=actor_path,
         )
+        self._container = VariableContainer("localhost", self._actor)
 
         # Critic network & target network
-        self._critic_1 = Critic(
+        self._critic = TwinCritic(
             state_shape=self._env.observation_space.shape,
             action_shape=self._env.action_space.shape,
             learning_rate=critic_learning_rate,
-            model_path=model_c1_path,
+            model_path=critic_path,
         )
-        self._critic_targ_1 = Critic(
+        self._critic_target = TwinCritic(
             state_shape=self._env.observation_space.shape,
             action_shape=self._env.action_space.shape,
-            learning_rate=critic_learning_rate,
-            model_path=model_c1_path,
         )
+        self._update_target(self._critic, self._critic_target, tau=tf.constant(1.0))
 
-        # Critic network & target network
-        self._critic_2 = Critic(
-            state_shape=self._env.observation_space.shape,
-            action_shape=self._env.action_space.shape,
-            learning_rate=critic_learning_rate,
-            model_path=model_c2_path,
-        )
-        self._critic_targ_2 = Critic(
-            state_shape=self._env.observation_space.shape,
-            action_shape=self._env.action_space.shape,
-            learning_rate=critic_learning_rate,
-            model_path=model_c2_path,
-        )
-
-        # first make a hard copy
-        self._update_target(self._critic_1, self._critic_targ_1, tau=tf.constant(1.0))
-        self._update_target(self._critic_2, self._critic_targ_2, tau=tf.constant(1.0))
-
+        # load db from checkpoint or make a new one
         if db_path is None:
             checkpointer = None
         else:
             checkpointer = reverb.checkpointers.DefaultCheckpointer(path=db_path)
 
-        # actual training step
-        self._train_step = tf.Variable(
-            0,
-            trainable=False,
-            dtype=tf.int32,
-            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
-            shape=(),
-        )
-
-        self._stop_agents = tf.Variable(
-            False,
-            trainable=False,
-            dtype=tf.bool,
-            aggregation=tf.VariableAggregation.ONLY_FIRST_REPLICA,
-            shape=(),
-        )
-
-        # prepare variable container
-        self._variables_container = {
-            "train_step": self._train_step,
-            "stop_agents": self._stop_agents,
-            "policy_variables": self._actor.model.variables,
-        }
-
-        # variables signature for variable container table
-        variable_container_signature = tf.nest.map_structure(
-            lambda variable: tf.TensorSpec(variable.shape, dtype=variable.dtype),
-            self._variables_container,
-        )
-
         # Initialize the reverb server
         self.server = reverb.Server(
             tables=[
@@ -158,7 +113,7 @@ class Learner:
                     name="experience",
                     sampler=reverb.selectors.Uniform(),
                     remover=reverb.selectors.Fifo(),
-                    rate_limiter=reverb.rate_limiters.MinSize(learning_starts),
+                    rate_limiter=reverb.rate_limiters.MinSize(warmup_steps),
                     max_size=buffer_capacity,
                     max_times_sampled=0,
                     signature={
@@ -185,22 +140,23 @@ class Learner:
                     rate_limiter=reverb.rate_limiters.MinSize(1),
                     max_size=1,
                     max_times_sampled=0,
-                    signature=variable_container_signature,
+                    signature=self._container.variable_container_signature,
                 ),
             ],
             port=8000,
             checkpointer=checkpointer,
         )
 
-        # Initializes the reverb client
+        # Initializes the reverb client and tf.dataset
         self.client = reverb.Client("localhost:8000")
-        self.tf_client = reverb.TFClient(server_address="localhost:8000")
         self.dataset_iterator = iter(
             reverb.TrajectoryDataset.from_table_signature(
                 server_address="localhost:8000",
                 table="experience",
                 max_in_flight_samples_per_worker=10,
-            ).batch(batch_size, drop_remainder=True)
+            )
+            .batch(batch_size, drop_remainder=True)
+            .prefetch(tf.data.experimental.AUTOTUNE)
         )
 
         # init Weights & Biases
@@ -209,7 +165,7 @@ class Learner:
 
             # Settings
             wandb.config.max_steps = max_steps
-            wandb.config.learning_starts = learning_starts
+            wandb.config.warmup_steps = warmup_steps
             wandb.config.buffer_capacity = buffer_capacity
             wandb.config.batch_size = batch_size
             wandb.config.actor_learning_rate = actor_learning_rate
@@ -219,7 +175,7 @@ class Learner:
             wandb.config.gamma = gamma
 
         # init actor's params in DB
-        self._push_variables()
+        self._container.push_variables()
 
     def _update_target(self, net, net_targ, tau):
         for source_weight, target_weight in zip(
@@ -227,65 +183,39 @@ class Learner:
         ):
             target_weight.assign(tau * source_weight + (1.0 - tau) * target_weight)
 
-    def _push_variables(self):
-        self.tf_client.insert(
-            data=tf.nest.flatten(self._variables_container),
-            tables=tf.constant(["variables"]),
-            priorities=tf.constant([1.0], dtype=tf.float64),
-        )
-
     # -------------------------------- update critic ------------------------------- #
     def _update_critic(self, batch):
         next_action, next_log_pi = self._actor.predict(batch.data["next_observation"])
 
         # target Q-values
-        next_q_1 = self._critic_targ_1.model(
+        next_q_value = self._critic_target.model(
             [batch.data["next_observation"], next_action]
         )
-        next_q_2 = self._critic_targ_2.model(
-            [batch.data["next_observation"], next_action]
-        )
-        next_q = tf.minimum(next_q_1, next_q_2)
 
         # Bellman Equation
-        Q_targets = tf.stop_gradient(
+        Q_target = tf.stop_gradient(
             batch.data["reward"]
             + (1 - batch.data["terminal"])
             * self._gamma
-            * (next_q - self._alpha * next_log_pi)
+            * (next_q_value - self._alpha * next_log_pi)
         )
 
-        # update critic '1'
+        # update critic
         with tf.GradientTape() as tape:
-            q_values = self._critic_1.model(
+            q_value = self._critic.model(
                 [batch.data["observation"], batch.data["action"]]
             )
             q_losses = tf.losses.huber(  # less sensitive to outliers in batch
-                y_true=Q_targets, y_pred=q_values
+                y_true=Q_target, y_pred=q_value
             )
-            q1_loss = tf.nn.compute_average_loss(q_losses)
+            q_loss = tf.nn.compute_average_loss(q_losses)
 
-        grads = tape.gradient(q1_loss, self._critic_1.model.trainable_variables)
-        self._critic_1.optimizer.apply_gradients(
-            zip(grads, self._critic_1.model.trainable_variables)
+        grads = tape.gradient(q_loss, self._critic.model.trainable_variables)
+        self._critic.optimizer.apply_gradients(
+            zip(grads, self._critic.model.trainable_variables)
         )
 
-        # update critic '2'
-        with tf.GradientTape() as tape:
-            q_values = self._critic_2.model(
-                [batch.data["observation"], batch.data["action"]]
-            )
-            q_losses = tf.losses.huber(  # less sensitive to outliers in batch
-                y_true=Q_targets, y_pred=q_values
-            )
-            q2_loss = tf.nn.compute_average_loss(q_losses)
-
-        grads = tape.gradient(q2_loss, self._critic_2.model.trainable_variables)
-        self._critic_2.optimizer.apply_gradients(
-            zip(grads, self._critic_2.model.trainable_variables)
-        )
-
-        return q1_loss + q2_loss
+        return q_loss
 
     # -------------------------------- update actor ------------------------------- #
     def _update_actor(self, batch):
@@ -294,11 +224,9 @@ class Learner:
             y_pred, log_pi = self._actor.predict(batch.data["observation"])
 
             # predict q value
-            q_1 = self._critic_1.model([batch.data["observation"], y_pred])
-            q_2 = self._critic_2.model([batch.data["observation"], y_pred])
-            q = tf.minimum(q_1, q_2)
+            q_value = self._critic.model([batch.data["observation"], y_pred])
 
-            policy_losses = self._alpha * log_pi - q
+            policy_losses = self._alpha * log_pi - q_value
             policy_loss = tf.nn.compute_average_loss(policy_losses)
 
         grads = tape.gradient(policy_loss, self._actor.model.trainable_variables)
@@ -342,18 +270,17 @@ class Learner:
         policy_loss = self._update_actor(sample)
 
         # Soft update target networks
-        self._update_target(self._critic_1, self._critic_targ_1, tau=self._tau)
-        self._update_target(self._critic_2, self._critic_targ_2, tau=self._tau)
+        self._update_target(self._critic, self._critic_target, tau=self._tau)
 
         # Store new actor's params
-        self._push_variables()
+        self._container.push_variables()
 
         return critic_loss, policy_loss, alpha_loss
 
     def run(self):
-        for train_step in range(self._learning_starts, self._max_steps):
+        for train_step in range(self._warmup_steps, self._max_steps):
             # update train_step (otlacok modelov)
-            self._train_step.assign(train_step)
+            self._container.train_step.assign(train_step)
 
             # update models
             critic_loss, policy_loss, alpha_loss = self._update()
@@ -366,8 +293,9 @@ class Learner:
                 print(f"Policy loss: {policy_loss}")
                 print("=============================================")
                 print(
-                    f"Training ... {tf.floor(self._train_step * 100 / self._max_steps)} %"  # noqa
+                    f"Training ... {tf.floor(train_step * 100 / self._max_steps)} %"  # noqa
                 )
+
             if self._log_wandb:
                 # log of epoch's mean loss
                 wandb.log(
@@ -381,15 +309,14 @@ class Learner:
                 )
 
         # Stop the agents
-        self._stop_agents.assign(True)
-        self._push_variables()
+        self._container.stop_agents.assign(True)
+        self._container.push_variables()
 
     def save(self):
         if self._save_path is not None:
             # store models
-            self._actor.model.save(os.path.join(self._save_path, "model_A.h5"))
-            self._critic_1.model.save(os.path.join(self._save_path, "model_C1.h5"))
-            self._critic_2.model.save(os.path.join(self._save_path, "model_C2.h5"))
+            self._actor.model.save(os.path.join(self._save_path, "actor"))
+            self._critic.model.save(os.path.join(self._save_path, "critic"))
 
         # store checkpoint of DB
         self.client.checkpoint()
